import * as express from 'express';
import * as cookieParser from 'cookie-parser';
import * as bodyParser from 'body-parser';
import { IOptions } from '../interface/options';
import { IReturn } from '../interface/return';
import { IRoute } from '../interface/route';
import { IContext } from '../interface/context';

export class Server {

	protected server: any;
	protected controllers: Array<any> = [];

	constructor(protected options: IOptions) {
		this.server = express();
		this.server.use(cookieParser());
		this.server.use(bodyParser.urlencoded({ extended: true }));
		this.server.use(bodyParser.json());
	}

	public async start(): Promise<void> {
		await this.processControllers();

		this.server.listen(this.options.port, () => {
			console.log('Listening to port: ' + this.options.port);
		});
	}

	private async processControllers(): Promise<void> {
		this.controllers.forEach((controller: any) => {

			// Get the controller information and meta.
			const instance: any = new controller();
			const prefix: string = Reflect.getMetadata('prefix', controller);
			const routes: Array<IRoute> = Reflect.getMetadata('routes', controller);

			// Loop the routes and setup the events.
			routes.forEach((route: any) => {
				this.server[route.requestMethod](prefix + route.path, async (request: express.Request, response: express.Response): Promise<void> => {

					// Define headers
					const headers: any = {};
					for (const index in request.headers) {
						headers[index] = request.headers[index];
					}

					// Define query
					const query: any = {};
					for (const index in request.query) {
						query[index] = request.query[index];
					}

					// Define params
					const params: any = {};
					for (const index in request.params) {
						params[index] = request.params[index];
					}

					// Define cookies
					const cookies: any = {};
					for (const index in request.cookies) {
						cookies[index] = request.cookies[index];
					}

					// Create the context object.
					const context: IContext = {
						httpVersion: request.httpVersion,
						headers: headers,
						requestedUrl: request.url,
						httpMethod: request.method,
						ipAddress: request.socket.remoteAddress || '',
						params: params,
						query: query,
						cookies: cookies,
						setCookie: (name: string, value: string, maxAge?: number, httpOnly?: boolean, secure?: boolean) => {
							const cookieParams: any = {};
							if (maxAge) cookieParams.maxAge = (1000 * 86400 * maxAge);
							if (httpOnly) cookieParams.httpOnly = httpOnly;
							if (secure) cookieParams.secure = secure;
							response.cookie(name, value, cookieParams);
							return true;
						}
					}

					// Call the function in the callback with the context.
					const result: IReturn = await instance[route.methodName](context);
					response.status(result.status).send(result.content);
				});
			});
		});
	}

	public registerControllers(controllers: Array<any>): void {
		for (const index in controllers) {
			this.controllers.push(controllers[index]);
		}
	}

	public registerMiddleware(middleware: (request: express.Request, response: express.Response, next: express.NextFunction) => void): void {
		this.server.use(middleware);
	}
}